/* BullDog Dynamic Firewall
   by Father Peter Darin

   Covered by GNU license
*/

#include "common.h"

/* Definitions */

//#define DEBUG
//#define BDEBUG

#define INFO                    "/sbin/bulldog/info"
#define INITFW                  "/sbin/bulldog/InitFW"
#define BLACKLIST               "/sbin/bulldog/blacklist"
#define STATE                   "/sbin/bulldog/state"
#define CONFIG                  "/sbin/bulldog/bulldog.conf"
#define MAXAD                   32
#define DELAY                   50
#define MAXPROC                 60
#define MINUTE                  60         // seconds
#define HOUR                    60*MINUTE
#define TIMEOUT                 (2*HOUR)
#define MAXIP                   1000

/* global varables */

typedef unsigned char               ip[16];  // ip address
typedef unsigned char               dn[64];  // domain name

/* General usage globals */

S
  {
    S ip ip;
    S tcphdr hdr;
    C filler[60-sizeof(struct ip)];
  } tcppkt;

S
  {
    S ip ip;
    S udphdr hdr;
    C filler[60-sizeof(struct ip)];
  } udppkt;

S
  {
    S ip ip;
    S icmphdr hdr;
    C filler[60-sizeof(struct ip)];
  } icmppkt;

dn BA[65536];  // Block addresses by partial domain, ie DoubleClick.net
I  LBA=0;      // Last entry

dn ER[65536];  // Allowed Echo request (ping) sites
I  LER=0;      // Last entry

C BR[65536];   // Block Ports absolutely
C SP[65536];   // Valid Source ports that do NOT get blocked

S
  {
    UI port;
    dn address;
  } AP[65536];  // Block port references that do not match these listed.
I LAP=0;

ip MAD[MAXAD];          // My addresses
I  LMA;                 // Last entry

S _ipstruct
  {
    ip ip;
    I  ttl;
  } BlackList;

S _ipstruct State;

FILE  *fd;

F proc1[MAXPROC];
I p1l=0, pfirst=TRUE;
D ptotal1=0;
F info[4], oldi[4], total=0.0, oldit=0.0;

I  oldsec, oldmin, oldhour, makelog=0;

I  prio=-20;

I  rawicmp;
I  rawtcp;
I  rawudp;

UL  tpc=0, upc=0, ipc=0;

/* Prototypes */

C *Resolver(C *resolved, C *sp);
I  IPBlocked(C *sp);
V  BlockAddress(C *sa, I sp, C *da, I dp);
V  Check(C *sa,I sp,C *da,I dp,I icmp);
V  CheckFilters(C *sa,I sp,C *da,I dp);
V  CheckHighPort(C *sa,I sp,C *da,I dp);
V  CheckICMPFilters(C *sa,I sp,C *da,I dp);
V  CheckRange(C *sa,I sp,C *da,I dp);
V  GetTCPdump(V);
V  LogPackets(C *sa,I sp,C *da,I dp,I it);
V  PassThru(I s);
V  ProcessPackets(V);
V  ReadConfig(V);
V  ReloadBlackList(V);
V  SigHandler(I sig);
V  GetProcessor(V);
V  Statistics(V);
V  Sweep(V);
V  StateSweep(V);
V  BalancePriority(V);
F  p(I x);
I  IsItMe(C *sa);
I  INState(C *sp);
V  AddState(C *sp, C *s);

/* Include code commons */

#include "dns.c"

I main(I argc, C *argv[])
  {
    STR    d1;
    time_t now;
    S      tm *tm;
        
    now=time(NULL);
    tm=localtime(&now);
    oldhour=tm->tm_hour;
    oldmin=tm->tm_min;
                            
    dstrcopy(logname,"/var/log/bulldog");
    
    signal(SIGHUP,  SigHandler);        // Logging
    signal(SIGUSR1,  SigHandler);       // Re-read Config file

    ReadConfig();

    rawicmp=socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    rawtcp=socket(AF_INET, SOCK_RAW, IPPROTO_TCP);
    rawudp=socket(AF_INET, SOCK_RAW, IPPROTO_UDP);

    PassThru(rawicmp);
    PassThru(rawtcp);
    PassThru(rawudp);

    printf("Bull Dog Activated...\n");
    writelog("Bull Dog Activated...");

    ReloadBlackList();

    close(0); /* Close stdin, stdout, stderr */
    close(1);
    close(2);
            
    if(!fork())
      {
        pid=getpid();
      
        fd=fopen("/var/run/bulldog.pid","w");
        fprintf(fd,"%i\n",pid);
        fclose(fd);
        sprintf(d1,"Bulldog running - Process ID=%i",pid);
        writelog(d1);

        while(1)
          ProcessPackets();
      }

    close(rawtcp);
    close(rawudp);
    close(rawicmp);

    return(0);
  }

/* Balance priority to not hog the CPU
*/

V BalancePriority(V)
  {
    STR d1;
    I p;

    p=(I)(ptotal1/2.5)-20;

    if(p>20)
      p=20;
    else
      if(p<-20)
        p=-20;  

    if(p<prio)          // We want to modify the priorty in
      prio--;           // in a consistant manner
    if(p>prio)
      prio++;  

    setpriority(PRIO_PROCESS,pid,prio);
  }
  
F p(I x)
  {
    F p=((info[x]-oldi[x])/(total-oldit))*100.0;
      
    return(p);
  }
  
V GetProcessor(V)
  {
    C str1[20];
    I i;

    fd=fopen("/proc/stat", "r");
    if(pfirst==TRUE)
      fscanf(fd, "%s %f %f %f %f", str1, oldi, oldi+1, oldi+2, oldi+3);
    else
      fscanf(fd, "%s %f %f %f %f", str1, info, info+1, info+2, info+3);
    fclose(fd);

    if(pfirst==TRUE)
      oldit=0.0;

    total=0.0;
    for(i=0; i<4; i++)
      {
        if(pfirst==TRUE)
          {
            oldi[i]/=2;
            oldit+=oldi[i];
            info[i]=oldi[i]*2;
          }
        total+=info[i];
      }

    proc1[p1l++]=p(0)+p(1)+p(2);

    if(p1l>MAXPROC-1)                p1l=0;

    ptotal1=0.0;

    if(pfirst==TRUE)
      {
        pfirst=FALSE;
        for(i=1;i<MAXPROC;i++)
          proc1[i]=proc1[0];
      }

    for(i=0;i<MAXPROC;i++)
      ptotal1+=proc1[i];

    ptotal1/=MAXPROC;

    if(ptotal1<0.0)                ptotal1=0.0;
    if(ptotal1>100.0)              ptotal1=100.0;

    oldit=total;
    for(i=0; i<4; i++)
      oldi[i]=info[i];
                  
  }

/* Display Statistics
*/

V Statistics(V)
  {
    STR d1;
    
    writelog("Packets Processed Last Hour by BullDog:");
    sprintf(d1,"  ->  TCP Packets: %i", tpc);
    writelog(d1);
    sprintf(d1,"  ->  UDP Packets: %i", upc);
    writelog(d1);
    sprintf(d1,"  -> ICMP Packets: %i", ipc);
    writelog(d1);
  
    ipc=0;
    tpc=0;
    upc=0;  
  }

/* Process the Packets
*/

V ProcessPackets(V)
  {
    time_t now;
    S      tm *tm;

    GetTCPdump();

    now=time(NULL);
    tm=localtime(&now);

    if(tm->tm_sec!=oldsec)
      {
        oldsec=tm->tm_sec;
        GetProcessor();
        Sweep();
        StateSweep();
      }    

    if(tm->tm_min!=oldmin)
      {
        oldmin=tm->tm_min;
        BalancePriority();
      }    

    if(tm->tm_hour!=oldhour)
      {
        oldhour=tm->tm_hour;
        Statistics();
      }    

    usleep(DELAY); 
  }

/* Log all packets
*/

V  LogPackets(C *sa,I sp,C *da,I dp,I it)
  {
    STR d1;

    if(makelog==1)
      {
        switch(it)
          {
            case -1:        // TCP packet
              sprintf(d1,"T %16s %5i %16s %5i",sa, sp, da, dp);
              fwritelog("/var/log/packets",d1);
              break;
            case -2:        // UDP packet
              sprintf(d1,"U %16s %5i %16s %5i",sa, sp, da, dp);
              fwritelog("/var/log/packets",d1);
              break;
            default:        // ICMP packet
              sprintf(d1,"I %16s %16s %5i",sa, da, it);
              fwritelog("/var/log/packets",d1);
              break;  
          }
      }
  }


/*  Check for absolute blocked ports
*/

V ReloadBlackList(V)
  {
    STR d1;
    I   fh1, i, lr;

    system(INITFW);
    writelog("Firewall Defaults Initializing...");

    fh1=sopen(BLACKLIST,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);
    lr=lseek(fh1,0,SEEK_END)/sizeof(BlackList);
    sprintf(d1,"Blocking %i stored entries",lr);
    writelog(d1);
    if(lr>0)
      {
        for(i=0;i<lr;i++)
          {
            lseek(fh1,i*sizeof(BlackList),SEEK_SET);
            read(fh1,&BlackList,sizeof(BlackList));
            
            sprintf(d1,"/sbin/bulldog/BlockIP %s",BlackList.ip);
            system(d1);
            sprintf(d1,"Blocked %s",BlackList.ip);
            writelog(d1);
          }
      }
    close(fh1);
  }

V CheckRange(C *sa,I sp,C *da,I dp)
  {
    if(BR[dp]=='Y')     // Blocked port ranges
      BlockAddress(sa,sp,da,dp);
  }

/*  This next procedure checks any high port values to determine if the 
    system is connected to a valid external service suchlike mail or
    ftp.
*/

V CheckHighPort(C *sa,I sp,C *da,I dp)
  {
    I i, d=FALSE;

    if(dp>1024) 
      {
        d=FALSE;        // Is the port one on the list ie extended servers?
        for(i=0;i<LAP && d==FALSE;i++)
          if(dp==AP[i].port)
            d=TRUE;    

        if(d==FALSE)  // Not in our partial list
          if(SP[sp]=='N')  // Is it an external service?
            BlockAddress(sa,sp,da,dp);
      }    
  }
  
/*  Compare partial source address to our filtered list.
*/

V CheckFilters(C *sa,I sp,C *da,I dp)
  {
    STR d1, resolved;
    I i, d=FALSE, e=FALSE;
    
    if(IPBlocked(sa)==FALSE)
      {
        Resolver(resolved,sa);

        if(resolved[0]!='\0')  // block addresses like DoubleClick
          {
            for(i=0;i<LBA;i++)
              if(pos(BA[i],resolved)>0 || dstrcmp(BA[i],resolved)==TRUE) 
                BlockAddress(sa,sp,da,dp);

            if(dp>0 && SP[sp]=='N')
              {
                d=FALSE;        // Is the port one on the list?
                for(i=0;i<LAP && d==FALSE;i++)
                  if(dp==AP[i].port)
                    d=TRUE;    

                if(d==TRUE)    // Yes it is
                  {
                    e=FALSE;
                    for(i=0;i<LAP && e==FALSE;i++)
                      {
                        if(dp==AP[i].port)  // Find the exact port
                          {  // Does the partial address and source match?
                            if(pos(AP[i].address,resolved)>0 
                            || dstrcmp(AP[i].address,resolved)==TRUE)
                              e=TRUE;  
                          }
                      }
                  }
                else  // open for now listed ports, we don't want 65,000 rules!
                  e=TRUE;
                  
                if(e==FALSE)  
                  BlockAddress(sa,sp,da,dp);
              }
          }
      }
  }  

/*  Check for sites that are allowed to ping this site
*/
  
V CheckICMPFilters(C *sa,I sp,C *da,I dp)
  {
    STR resolved;
    I   i, d=FALSE;
    
    if(IPBlocked(sa)==FALSE)
      {
        Resolver(resolved,sa);

        if(resolved[0]!='\0')
          {
            for(i=0;i<LER && d==FALSE;i++)
              if(pos(ER[i],resolved)>0 || dstrcmp(ER[i],resolved)==TRUE)
                d=TRUE;  // Source is a valid ping site
              
            if(d==FALSE)  
              BlockAddress(sa,sp,da,dp);
          }      
      }          
  }
  
I IsItMe(C *sa)
  {
    I i,done=FALSE;
    
    for(i=0;i<LMA && done==FALSE;i++)
      done=dstrcmp(MAD[i],sa);
      
    return(done);  
  }  
  
/*  Only process if WE are the destination and we are NOT the source
*/

V Check(C *sa,I sp,C *da,I dp,I icmp)
  {
    if(IsItMe(da)==TRUE)   // Is the destination me?
      {
        if(IsItMe(sa)==FALSE)   // Is this from me?
          {
            CheckRange(sa,sp,da,dp);          

            CheckHighPort(sa,sp,da,dp);  

            if(icmp==TRUE)
              CheckICMPFilters(sa,sp,da,dp);

            CheckFilters(sa,sp,da,dp);
          }
      }
  }

V Sweep(V)
  {
    STR d1;
    I   fh1, i, lr;

    fh1=sopen(BLACKLIST,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);
    lr=lseek(fh1,0,SEEK_END)/sizeof(BlackList);
    if(lr>0)
      {
        for(i=0;i<lr;i++)
          {
            lseek(fh1,i*sizeof(BlackList),SEEK_SET);
            read(fh1,&BlackList,sizeof(BlackList));
            
            if(BlackList.ttl>0)
              {
                BlackList.ttl--;
                lseek(fh1,i*sizeof(BlackList),SEEK_SET);
                write(fh1,&BlackList,sizeof(BlackList));
              }
            else
              {
                sprintf(d1,"/sbin/bulldog/UnBlockIP %s &",BlackList.ip);
                system(d1);
                
                sprintf(d1,"Unblocked %s",BlackList.ip);
                writelog(d1);

                if(i<lr-1)
                  {
                    lseek(fh1,(lr-1)*sizeof(BlackList),SEEK_SET);
                    read(fh1,&BlackList,sizeof(BlackList));

                    lseek(fh1,i*sizeof(BlackList),SEEK_SET);
                    write(fh1,&BlackList,sizeof(BlackList));
                  }

                lr--;
                ftruncate(fh1,sizeof(BlackList)*lr);
              }  
          }
      }
    close(fh1);
  }

V StateSweep(V)
  {
    STR d1;
    I   fh1, i, lr;

    fh1=sopen(STATE,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);
    lr=lseek(fh1,0,SEEK_END)/sizeof(State);
    if(lr>0)
      {
        for(i=0;i<lr;i++)
          {
            lseek(fh1,i*sizeof(State),SEEK_SET);
            read(fh1,&State,sizeof(State));
            
            if(State.ttl>0)
              {
                State.ttl--;
                lseek(fh1,i*sizeof(State),SEEK_SET);
                write(fh1,&State,sizeof(State));
              }
            else
              {
                if(i<lr-1)
                  {
                    lseek(fh1,(lr-1)*sizeof(State),SEEK_SET);
                    read(fh1,&State,sizeof(State));

                    lseek(fh1,i*sizeof(State),SEEK_SET);
                    write(fh1,&State,sizeof(State));
                  }

                lr--;
                ftruncate(fh1,sizeof(State)*lr);
              }  
          }
      }
    close(fh1);
  }

V BlockAddress(C *sa, I sp, C *da, I dp)
  {
    STR    d1, resolved;
    I      fh1, i, done=FALSE;

    done=IPBlocked(sa);

    if(done==FALSE)
      {
        dstrcopy(BlackList.ip,sa);
        BlackList.ttl=TIMEOUT;
    
        fh1=sopen(BLACKLIST,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);
        lseek(fh1,0,SEEK_END);
        write(fh1,&BlackList,sizeof(BlackList));
        close(fh1);
      
        sprintf(d1,"/sbin/bulldog/BlockIP %s &",sa);
        system(d1);
        
        Resolver(resolved,sa);

        sprintf(d1,"Blocked %s(%i) -> %s(%i)",sa, sp, da, dp);
        writelog(d1);
        
        if(resolved[0]!='\0')
          {
            sprintf(d1,"        -> %s",resolved);
            writelog(d1);
          }    
        
#ifdef BDEBUG
        printf("B %16s %5i %16s %5i\n",sa, sp, da, dp);
#endif            
      }
  }

/* We use this to keep track of recent IP addresses so we don't over   
   run our process list.  DNS-Dog is spawened.  Without this, BullDog
   spawns DNS-Dog hundreds of times with the same IP address.
*/   

V AddState(C *sp, C *s)
  {
    I     fh1;

    dstrcopy(State.ip,sp);
    State.ttl=120;
    
    fh1=sopen(STATE,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);
    lseek(fh1,0,SEEK_END);
    write(fh1,&State,sizeof(State));
    close(fh1);
      
    system(s);
  }

V GetTCPdump(V)
  {
    ip sa, da;
    UI sp,  dp;    
    STR d1;
    I   a;
    I   size, icmp;

    size=read(rawicmp, &icmppkt, sizeof(icmppkt));
    if(size>=sizeof(icmppkt.ip) && size!=-1)
      {
        icmp=FALSE;
        
        a=icmppkt.ip.ip_src.s_addr;
        sprintf(sa,"%i.%i.%i.%i",a&255,a>>8&255,a>>16&255,a>>24&255);
        sp=0;
        
        a=icmppkt.ip.ip_dst.s_addr;
        sprintf(da,"%i.%i.%i.%i",a&255,a>>8&255,a>>16&255,a>>24&255);
        dp=0;

#ifdef DEBUG
        printf("I %16s %5i %16s %5i %i\n",sa, sp, da, dp, icmppkt.hdr.type);
#endif            

        if(INDNSList(sa)==FALSE && INState(sa)==FALSE)
          AddState(sa,dstrcat(d1,"/sbin/bulldog/dns-dog ",sa," &",NULL));

/* is it a valid ICMP type?  */

        if(icmppkt.hdr.type!=ICMP_DEST_UNREACH
        && icmppkt.hdr.type!=ICMP_SOURCE_QUENCH
        && icmppkt.hdr.type!=ICMP_TIME_EXCEEDED
        && icmppkt.hdr.type!=ICMP_PARAMETERPROB)
          icmp=TRUE;

        ipc++;

        Check(sa,sp,da,dp,icmp);
        LogPackets(sa,sp,da,dp,icmppkt.hdr.type);
      }
            
    size=read(rawtcp, &tcppkt, sizeof(tcppkt));
    if(size>=sizeof(tcppkt.ip) && size!=-1)
      {
        a=tcppkt.ip.ip_src.s_addr;
        sprintf(sa,"%i.%i.%i.%i",a&255,a>>8&255,a>>16&255,a>>24&255);
        sp=ntohs(tcppkt.hdr.th_sport);
        
        a=tcppkt.ip.ip_dst.s_addr;
        sprintf(da,"%i.%i.%i.%i",a&255,a>>8&255,a>>16&255,a>>24&255);
        dp=ntohs(tcppkt.hdr.th_dport);

#ifdef DEBUG
        printf("T %16s %5i %16s %5i\n",sa, sp, da, dp);
#endif            
        if(INDNSList(sa)==FALSE && INState(sa)==FALSE)
          AddState(sa,dstrcat(d1,"/sbin/bulldog/dns-dog ",sa," &",NULL));

        tpc++;

        Check(sa,sp,da,dp,FALSE);
        LogPackets(sa,sp,da,dp,-1);
      }
                                                                                    
    size=read(rawudp, &udppkt, sizeof(udppkt));
    if(size>=sizeof(udppkt.ip) && size!=-1)
      {
        a=udppkt.ip.ip_src.s_addr;
        sprintf(sa,"%i.%i.%i.%i",a&255,a>>8&255,a>>16&255,a>>24&255);
        sp=ntohs(udppkt.hdr.uh_sport);
        
        a=udppkt.ip.ip_dst.s_addr;
        sprintf(da,"%i.%i.%i.%i",a&255,a>>8&255,a>>16&255,a>>24&255);
        dp=ntohs(udppkt.hdr.uh_dport);

#ifdef DEBUG
        printf("U %16s %5i %16s %5i\n",sa, sp, da, dp);
#endif            
        if(INDNSList(sa)==FALSE && INState(sa)==FALSE)
          AddState(sa,dstrcat(d1,"/sbin/bulldog/dns-dog ",sa," &",NULL));

        upc++;

        Check(sa,sp,da,dp,FALSE);
        LogPackets(sa,sp,da,dp,-2);
      }
  }

V ReadConfig(V)
  {
    STR d1, d2;
    I   i, s, e;

    memset(BR,'N',sizeof(BR));
    memset(SP,'N',sizeof(SP));
    
    fd=fopen(CONFIG,"r");
    while(fgets(d1,sizeof(d1),fd)!=NULL)
      {
        trim(d1,d1);
        if(d1[0]=='#')  // ignore comments
          continue;
        d1[dstrlen(d1)-1]='\0'; // strip newline
        trim(d1,d1);
        if(d1[0]=='\0')  // ignore blank lines
          continue; 
        switch(d1[0])
          {
            case 'a':
            case 'A':
              switch(d1[1])
                {
                  case 'p':  // port
                  case 'P':
                    for(i=dstrlen(d1);d1[i]!=' ';i--);
                    substr(AP[LAP].address,d1,i+2,dstrlen(d1));
                    d1[i]='\0';
                    d1[0]=' ';
                    d1[1]=' ';
                    ltrim(d1,d1);       // port number
                    AP[LAP++].port=s2l(d1);
                    break;
                }    
              break;    // Break for A
            case 'b':  // block address
            case 'B':
              switch(d1[1])
                {
                  case 'a':  // partial address
                  case 'A':
                    d1[0]=' ';
                    d1[1]=' ';
                    ltrim(BA[LBA++],d1);
                    break;
                  case 'r':  // Range of ports
                  case 'R':
                    d1[0]=' ';
                    d1[1]=' ';
                    ltrim(d1,d1);
                    i=pos(" ",d1);
                    substr(d2,d1,i+1,dstrlen(d1));
                    ltrim(d2,d2);
                    e=s2l(d2);
                    d1[i]='\0';
                    s=s2l(d1);
                    for(i=s;i<=e;i++)
                      BR[i]='Y';
                    break;
                  case 'p':  // Single port
                  case 'P':
                    d1[0]=' ';
                    d1[1]=' ';
                    ltrim(d1,d1);
                    s=s2l(d1);
                    BR[s]='Y';
                    break;
                }
              break;  // Break for B
           case 'i':
           case 'I':
             d1[0]=' ';
             ltrim(ER[LER++],d1);
             break;   // Break for I      
           case 's':
           case 'S': 
              switch(d1[1])
                {
                  case 'p':  // port number
                  case 'P':
                    d1[0]=' ';
                    d1[1]=' ';
                    ltrim(d1,d1);
                    s=s2l(d1);
                    SP[s]='Y';
                    break;
                }  
              break;  // Break for S
           case 'm':
           case 'M': 
              switch(d1[1])
                {
                  case 'a':  // port number
                  case 'A':
                   d1[0]=' ';
                   d1[1]=' ';
                   ltrim(MAD[LMA++],d1);
                    break;
                }  
              break;  // Break for S
          }  
      }
    fclose(fd);
  }
  
I IPBlocked(C *sp)
  {
    I fh1, i, lr, done=FALSE;

    fh1=sopen(BLACKLIST,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);
    lr=lseek(fh1,0,SEEK_END)/sizeof(BlackList);
    for(i=0;i<lr && done==FALSE;i++)
      {
        lseek(fh1,i*sizeof(BlackList),SEEK_SET);
        read(fh1,&BlackList,sizeof(BlackList));

        done=dstrcmp(sp,BlackList.ip);
      }
    close(fh1);
    
    return(done);
  }

C *Resolver(C *resolved, C *sp)
  {
    S
      {
        ip ip;
        dn resolved;
      } DNSList;

    dn d1, d2;
    I  fh1, i, lr, done=FALSE;

    dstrcopy(d1,sp);

    for(i=dstrlen(d1);d1[i]!='.';i--);
    d1[i]='\0';
    for(i=0;d1[i]!='\0';i++)
      if(d1[i]=='.')
        d1[i]='/';
            
    sprintf(d2,"%s%s",DNSLIST,d1);

    if(exist(d2)==FALSE)
      {
        resolved[0]='\0';                            
        return(resolved);
      }
                
    fh1=sopen(d2,O_RDONLY,SH_DENYNO,S_IREAD);
    if(fh1>0)
      {
        lr=lseek(fh1,0,SEEK_END)/sizeof(DNSList);
        for(i=0;i<lr && done==FALSE;i++)
          {
            lseek(fh1,i*sizeof(DNSList),SEEK_SET);
            read(fh1,&DNSList,sizeof(DNSList));
    
            done=dstrcmp(sp,DNSList.ip);
          }
        close(fh1);
      }

    if(done==TRUE)
      dstrcopy(resolved,DNSList.resolved);
    else
      resolved[0]='\0';                            

#ifdef DEBUG    
    printf("R: %s\n",resolved);
#endif
    return(resolved);
  }
  
V PassThru(I s) 
  {
    I opts;
    
    opts=fcntl(s,F_GETFL);
    opts=(opts|O_NONBLOCK);
    fcntl(s,F_SETFL,opts);
  }

V SigHandler(I sig)
  {
    switch(sig)
      {
        case SIGHUP:
          makelog=1-makelog;
          switch(makelog)
            {
              case 0:
                writelog("Packet Logging OFF");
                break;
              case 1:
                writelog("Packet Logging ON");
                break;
            }
        case SIGUSR1:
          writelog("Reloading configuration...");
          ReadConfig();
          break;    
        default:
            printf("burp\n");
            break;
      }
  }
                
I INState(C *sp)
  {
    I fh1, i, lr, done=FALSE;

    fh1=sopen(STATE,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);
    lr=lseek(fh1,0,SEEK_END)/sizeof(State);
    for(i=0;i<lr && done==FALSE;i++)
      {
        lseek(fh1,i*sizeof(State),SEEK_SET);
        read(fh1,&State,sizeof(State));

        done=dstrcmp(sp,State.ip);
      }
    close(fh1);

    return(done);
  }

#include "common.c"
