/* DNS Dog - DNS tool to speed up main firewall inquiries
   part of BullDog Dynamic Firewall
   by Father Peter Darin

   Covered by GNU license
*/

#include "common.h"

/* Definitions */

//#define DEBUG

/* global varables */

typedef unsigned char               ip[16];  // ip address
typedef unsigned char               dn[64];  // domain name

/* General usage globals */

I  fh1;
L  DNSloc=0, count=0;
I  oldhour=-1;

/* Prototypes */

C *Resolver(C *d, C *sp);
V  AddAddress(C *sp);
V  DoResolve(C *f);
V  FindDB(C *d);

/* Include code commons */

#include "dns.c"  

I main(I argc, C *argv[])
  {
    FILE   *fd;
    STR     d1;
        
    if(argc>1)
      {
        AddAddress(argv[1]);  // Nothing should call this except BullDog
        exit(0);
      }

    dstrcopy(logname,"/var/log/bulldog");
                        
    printf("DNS Dog Activated...\n");

    close(0); /* Close stdin, stdout, stderr */
    close(1);
    close(2);
            
    if(!fork())
      {
        pid=getpid();
        
        fd=fopen("/var/run/dns-dog.pid","w");
        fprintf(fd,"%i\n",pid);
        fclose(fd);
        sprintf(d1,"DNS dog running - Process ID=%i",pid);
        writelog(d1);

        while(1)
          {
            count=0;

            FindDB("/sbin/bulldog/dns");
          
            if(count>0)
              {
                sprintf(d1,"DNS entries: %i",count);
                writelog(d1);
              }    

            sleep(1);
          }
      }

    return(0);
  }

V FindDB(C *d)
  {
    S       stat   statf;
    S       dirent *dp;
    DIR    *dfd;    
    STR     d1;

    dfd=opendir(d);
  
    while((dp=readdir(dfd))!=NULL)
      {
        if(dp->d_name[0]!='.')
          {    
            sprintf(d1,"%s/%s",d,dp->d_name);
            stat(d1,&statf);
            if((statf.st_mode&S_IFDIR)==S_IFDIR)
              FindDB(d1);
            else  
              DoResolve(d1);
          }    
      }
    closedir(dfd);
  }

V DoResolve(C *f)
  {
    STR    r, d1;
    I      lr, i;
    time_t now;
    S      tm *tm;
                    
    fh1=sopen(f,O_RDWR,SH_DENYNO,S_IREAD|S_IWRITE);

    lr=lseek(fh1,0,SEEK_END)/sizeof(DNSList);

    count+=lr;

    if(lr>0)
      {
        for(DNSloc=0;DNSloc<lr;DNSloc++)
          {
            lseek(fh1,DNSloc*sizeof(DNSList),SEEK_SET);
            read(fh1,&DNSList,sizeof(DNSList));
    
            Resolver(DNSList.resolved,DNSList.ip);
            lseek(fh1,DNSloc*sizeof(DNSList),SEEK_SET);
            write(fh1,&DNSList,sizeof(DNSList));

            sleep(1);  
          }
        close(fh1);
      }
    else
      {
        close(fh1);
        remove(f);
      }  
  }

V AddAddress(C *sp)
  {
    dn d1, d2, d3;
    I  i, j, done=FALSE;

    done=INDNSList(sp);

    if(done==FALSE)
      {
        dstrcopy(DNSList.ip,sp);
        DNSList.resolved[0]='\0';
    
        Resolver(DNSList.resolved,DNSList.ip);
        if(DNSList.resolved[0]!='\0') // Only add if resolvable
          {
            dstrcopy(d1,sp);
            
            for(i=dstrlen(d1);d1[i]!='.';i--);
            d1[i]='\0';
            for(i=0;d1[i]!='\0';i++)
              if(d1[i]=='.')
                d1[i]='/';
            
            //   /usr/sbin/dns/207
            
            for(i=0;d1[i]!='/';i++)
              d2[i]=d1[i];
            d2[i]='\0';  
            sprintf(d3,"%s%s",DNSLIST,d2);
            if(exist(d3)==FALSE)
              mkdir(d3,S_IREAD|S_IWRITE);

            //   /usr/sbin/207/232
            
            for(j=i+1;d1[j]!='/';j++);
            for(i=0;i<j;i++)
              d2[i]=d1[i];
            d2[i]='\0';  
            sprintf(d3,"%s%s",DNSLIST,d2);
            if(exist(d3)==FALSE)
              mkdir(d3,S_IREAD|S_IWRITE);

            // Make DNS extry

            sprintf(d2,"%s%s",DNSLIST,d1);      
            fh1=sopen(d2,O_RDWR|O_CREAT,SH_DENYNO,S_IREAD|S_IWRITE);
            lseek(fh1,0,SEEK_END);
            write(fh1,&DNSList,sizeof(DNSList));
            close(fh1);
          }
      }
  }

C *Resolver(C *d, C *sp)
  {
    S   hostent *hent;
    I   addr;
    C   ia[4];

    dupstr(d,'\0',sizeof(dn));
    
    addr=inet_addr(sp);
    ia[3] = addr >> 24 & 255;
    ia[2] = addr >> 16 & 255;
    ia[1] = addr >>  8 & 255;
    ia[0] = addr       & 255;
    hent = gethostbyaddr (ia, 4, AF_INET);
    if(hent && hent->h_name && *hent->h_name)
      dstrcopy(d,hent->h_name);
    else
      d[0]='\0';
      
    return(d);
  }

#include "common.c"